# -*- coding: utf-8 -*-
# (c) 2018 The mqttwarn developers

# Configuration- and function files used by the test harness
configfile = 'tests/selftest.ini'
funcfile = 'tests/selftest.py'
bad_funcfile = 'tests/bad_funcs.py'
